from bingproxy.proxy import InvocationHandler
from bingdog.appconfig import ConfiguredTaskUtil, Configurator
from bingdog.logger import Logger

class MappedInvocationHandler(InvocationHandler):
    def __init__(self):
        self._taskMap = {}
        self._taskObjMap = {}
        
    def _fetchTaskMap(self):
        if len(self._taskMap) == 0:
            self._taskMap = Configurator.configuration['taskMap']

    def _getTaskHandler(self, proxy, nestedObj, *args, **kwargs):
        if self._taskObjMap.get(proxy) is None:
            handlerClass = self._taskMap.get(nestedObj.__module__ + "." + nestedObj.__class__.__name__)
            if handlerClass is None:
                return None
            else:
                self._taskObjMap[proxy] = handlerClass(nestedObj, *args, **kwargs)
        return self._taskObjMap[proxy]
    
    def invoke(self, proxy, func, nestedObj, *args, **kwargs):
        self._fetchTaskMap()
        taskHandler = self._getTaskHandler(proxy, nestedObj, *args, **kwargs)
        if (taskHandler):
            return getattr(taskHandler, func.__name__)(*args, **kwargs)
        else:
            return func(*args, **kwargs)

class FlowedInvocationHandler(InvocationHandler):
    def __init__(self):
        super().__init__()
        self.__configuredUtil = ConfiguredTaskUtil(Configurator.flowDiagramPath)
        self._taskObjMap = {}
        
    def invoke(self, proxy, func, nestedObj, *args, **kwargs):
        taskHandler = self._getTaskHandler(proxy, nestedObj, *args, **kwargs)
        if (taskHandler):
            return getattr(taskHandler, func.__name__)(*args, **kwargs)
        else:
            return func(*args, **kwargs)

    def _getTaskHandler(self, proxy, nestedObj, *args, **kwargs):
        if self._taskObjMap.get(proxy) is None:
            Logger.getLogger(Logger).info("Class Initialized: " + nestedObj.__class__.__name__)
            handlerClass = self.__configuredUtil.getTaskHandlerClass(nestedObj.taskId)
            if handlerClass is None:
                return None
            else:
                self._taskObjMap[proxy] = handlerClass(nestedObj, self.__configuredUtil, *args, **kwargs)
        return self._taskObjMap[proxy]


import logging
from bingdog.util import equalsIgnoreCase
from bingdog.appconfig import Configurator

class Logger(object):
    __logger = None

    def __init__(self):
        loggingLevel = None
        configurator = Configurator.getConfigurator()
        if equalsIgnoreCase("info", configurator.getLoggingLevel()):
            loggingLevel = logging.INFO
        elif equalsIgnoreCase("debug", configurator.getLoggingLevel()):
            loggingLevel = logging.DEBUG
        logging.basicConfig(level=loggingLevel, filename=configurator.getLoggingFilePath(), format=configurator.getLoggingFormat())
        
    def info(self, message):
        logging.info(message)
    
    def warning(self, message):
        logging.warning(message)
    
    def error(self, message):
        logging.error(message)
    
    def exception(self, message):
        logging.exception(message)
    
    def getLogger(loggerClass, *args, **kargs):
        if issubclass(loggerClass, Logger) or loggerClass is Logger:
            if __logger is None:
                __logger = loggerClass(*args, **kargs)
        else:
            raise LoggerClassException(loggerClass)
        return __logger
        
class LoggerClassException(Exception):
    
    def __init__(self, loggerClass):
        super(HandlerException, self).__init__(loggerClass, " is not a class of Logger.")


from abc import abstractmethod
from bingdog.util import ExtProcessShellUtil, ifNone, NullPointerException
from bingproxy.proxy import ProxyDecorator
from bingdog.taskproxy import FlowedInvocationHandler
class Task(object):
    
    def __init__(self, taskId):
        self.params = dict()
        self._taskId = taskId
    
    @property
    def taskId(self):
        return self._taskId
    
    @abstractmethod
    def run(self):
        pass

    def getNextTask(self):
        return None
        
    def hasNextChild(self):
        return False
    
    def getNextChild(self):
        return None
        
    def _processStatement(self,statement):
        if (statement):
            for key in self.params:
                statement = statement.replace("$" + str(key), str(self.params[key]))
            return statement
        else:
            return None

@ProxyDecorator(FlowedInvocationHandler)
class ShellExecutionTask(Task):
    def __init__(self, taskId):
        super().__init__(taskId)
        self.statement = None
        self._extUtil = ExtProcessShellUtil()
    
    def run(self):
        try:
            self._extUtil.execute(self._processStatement(ifNone(self.statement)))
        except NullPointerException as e:
            raise TaskExecutionException("None statement")

class TaskExecutionException(Exception):
    def __init__(self, message, exception = None):
        super().__init__(message + " has thrown an execution exception.")
        self.__nestedException = exception



import subprocess
from abc import abstractmethod
import logging

class ExtExecutionException(Exception):
    
    def __init__(self):
        super().__init__("External execution error.")

class ExtProcessUtil(object):

    @abstractmethod
    def execute(self, statement):
        pass

class ExtProcessShellUtil(ExtProcessUtil):

    def execute(self, statement):
        self.__print("Executing: " + statement)
        statusCode, responseText = subprocess.getstatusoutput(statement)
        if statusCode == -1:
            raise ExtExecutionException() 
        else:
            self.__print("Response Text: " + responseText)
        return responseText

    def __print(self, text):
        if len(text) <= 500:
            logging.info(text)
            
def ifNone(obj) :
    if obj is None:
        raise NullPointerException()
    else:
        return obj
    
class NullPointerException(Exception):
    pass
    
def equalsIgnoreCase(str1, str2):
    if str1 is None or str2 is None:
        return False
    str1 = str1.upper()
    str2 = str2.upper()
    return str1 == str2


import json
from bingdog.util import ifNone, NullPointerException
from bingdog.taskhandler import ConfiguredTaskHandler
import sys

class ConfiguredTaskUtil(object):
    def __init__(self, filePath):
        with open(filePath, "r") as conf:
            self._configJson = json.loads(conf.read())
            classPath = self._configJson.get("class_path")
            if (classPath) :
                for i in range(len(classPath.split(","))):
                    if (classPath.split(",")[i]):
                        sys.path.append(classPath.split(",")[i])
            
    def __getModuleClassName(self, fullName):
        if len(fullName.split("|")) < 2:
            raise Task.TaskExecutionException("Invalid Class Name")
        else:
            objModule = fullName.split("|")[0]
            objName = fullName.split("|")[1]
        return objModule, objName
    
    def _getClassByName(self, fullName):
        objModule, objName = self.__getModuleClassName(fullName)
        tokens = objName.split(".")
        module = __import__(objModule)
        for token in tokens:
            module = getattr(module, token)
        return module
    
    def _getTaskHandlerName(self, taskId):
        try:
            subTaskJson = ifNone(self._getSubTasksJson(taskId))
            return ifNone(subTaskJson.get('handler'))
        except NullPointerException as e:
            return None

    def getTask(self, taskId):
        task = None
        try:
            className = self._getTaskClassName(taskId)
            taskClass = self._getClassByName(ifNone(className))
            task = ifNone(taskClass)().newInstance(taskId)
            task.statement = self._getStatement(taskId)
        except NullPointerException as e:
            raise TaskExecutionException("Task ID: " + taskId + ". Incorrect Class Name", e)
        finally:
            return task

    def getNextTaskId(self, taskId):
        try:
            return ifNone(self._getTaskJson(taskId)).get('next_task')
        except NullPointerException as e:
            return None
    
    def getSubTasksJsonList(self, taskId):
        try:
            return ifNone(self._getSubTasksJson(taskId)).get('list')
        except NullPointerException as e:
            return None
    
    def getSubUnitTaskId(self, taskId):
        try:
            return ifNone(self._getSubTasksJson(taskId)).get("unit_task")
        except NullPointerException as e:
            return None
    
    def getSubListParamKey(self, taskId):
        try:
            return ifNone(self._getSubTasksJson(taskId)).get("list_param_key")
        except NullPointerException as e:
            return None 
    
    def getSubUnitParamKey(self, taskId):
        try:
            return ifNone(self._getSubTasksJson(taskId)).get("unit_param_key")
        except NullPointerException as e:
            return None

    def _getStatement(self, taskId):
        try:
            return ifNone(self._getTaskJson(taskId)).get("statement")
        except NullPointerException as e:
            return None

    def getTaskHandlerClass(self, taskId):
        try:
            return ifNone(self._getClassByName(ifNone(self._getTaskHandlerName(taskId))))
        except NullPointerException as e:
            return ConfiguredTaskHandler

    def _getTaskClassName(self, taskId):
        try:
            return ifNone(self._getTaskJson(taskId)).get("class_name")
        except NullPointerException as e:
            return None

    def _getTaskJson(self, taskId):
        try:
            return ifNone(self._configJson).get(taskId)
        except NullPointerException as e:
            return None
            
    def _getSubTasksJson(self, taskId):
        try:
            return ifNone(self._getTaskJson(taskId)).get('sub_task_list')
        except NullPointerException as e:
            return None


from bingdog.task import TaskExecutionException
from bingdog.appconfig import Configurator, ConfiguredTaskUtil
from bingdog.logger import Logger

class TaskExecutor():
        
    def execute(self,taskId):
        try:
            taskUtil = ConfiguredTaskUtil(Configurator.getConfigurator().flowDiagramPath)
            task = taskUtil.getTask(taskId)
            if (task):
                self.__execute(task)
            else:
                raise TaskExecutionException("None Root Task")
        except TaskExecutionException as e:
            logging.exception("Root TaskExecutionException has Thrown.")
        
    def __execute(self, task):
        if (task):
            try:
                task.run()
                while task.hasNextChild():
                    childTask = task.getNextChild()
                    self.__execute(childTask)
                nextTask = task.getNextTask()
                if (nextTask):
                    self.__execute(nextTask)
            except TaskExecutionException as e:
                logging.exception("TaskExecutionException has Thrown.")



from abc import abstractmethod

from bingproxy.proxy import InvocationHandler
from bingdog.util import ifNone, NullPointerException
import json
        
class TaskHandler(object):

    def __init__(self, nestedObj):
        super().__init__()
        self._childIndex = 0
        self._nestedObj = nestedObj
        
    def run(self):
        self._nestedObj.run()
        
    def getNextTask(self):
        task = self._getNextTask()
        if (task):
            task.params.update(self._nestedObj.params)
        return task
        
    def _getNextTask(self):
        return None
        
    def hasNextChild(self):
        if self._childIndex < self._getSubTaskListSize():
            return True
        else:
            return False
    
    def getNextChild(self):
        subTask = self._fetchNextSubTask()
        subTask.params.update(self._nestedObj.params)
        self._childIndex = self._childIndex + 1
        return subTask
    
    def _fetchNextSubTask(self):
        return None
    
    def _getSubTaskListSize(self):
        return 0

class ConfiguredTaskHandler(TaskHandler):
    def __init__(self, nestedObj, configuredUtil):
        super().__init__(nestedObj)
        self._configuredUtil = configuredUtil
    
    def _getNextTask(self):
        try:
            return ifNone(self._configuredUtil.getTask(ifNone(self._configuredUtil.getNextTaskId(self._nestedObj.taskId))))
        except NullPointerException as e:
            return None
    
    def hasNextChild(self):
        if self._childIndex < self._getSubTaskListSize():
            return True
        else:
            return False
    
    def _fetchNextSubTask(self):
        subTaskJsonList = self._configuredUtil.getSubTasksJsonList(self._nestedObj.taskId)
        if (subTaskJsonList):
            return self._configuredUtil.getTask(subTaskJsonList[self._childIndex])
        else:
            return self._configuredUtil.getTask(self._configuredUtil.getSubUnitTaskId(self._nestedObj.taskId))
    
    def getNextChild(self):
        subTask = self._fetchNextSubTask()
        try:
            self._nestedObj.params[ifNone(self._configuredUtil.getSubUnitParamKey(self._nestedObj.taskId))] = self._nestedObj.params[ifNone(self._configuredUtil.getSubListParamKey(self._nestedObj.taskId))][self._childIndex]
        finally:
            subTask.params.update(self._nestedObj.params)
            self._childIndex = self._childIndex + 1
            return subTask
    
    def _getSubTaskListSize(self):
        try:
            jsonList = self._configuredUtil.getSubTasksJsonList(self._nestedObj.taskId)
            return len(ifNone(jsonList))
        except NullPointerException as e:
            try:
                listKey = self._configuredUtil.getSubListParamKey(self._nestedObj.taskId)
                return len(self._nestedObj.params[ifNone(listKey)])
            except NullPointerException as e:
                return 0



import yaml
from bingdog.util import ifNone, NullPointerException

class Configurator(object):
    __configurator = None
    
    def __init__(self, confPath = "bingdog.conf"):
        self._config = None
        self.__flowDiagramPath = None
        with open(confPath, "r", "utf-8") as f:
            self._config = yaml.load(f)
    
    def initialize(confClass, *args, **kwargs):
        if issubclass(confClass, Configurator) or confClass is Configurator:
            if __configurator == None:
                __configurator = confClass(*args, **kwargs)
        else:
            raise ConfiguratorException(confClass)
    
    def getConfigurator():
        return ifNone(__configurator)
    
    @property
    def flowDiagramPath(self):
        if self.__flowDiagramPath is None:
            application = self._config.get("application")
            task = ifNone(application).get("task")
            self.__flowDiagramPath = ifNone(task).get("flow_file_path")
        return self.__flowDiagramPath
        
    @flowDiagramPath.setter    
    def flowDiagramPath(self, path):
        self.__flowDiagramPath = path
    
    def getLoggingLevel(self):
        try:
            application = self._config.get("application")
            logger = ifNone(application).get("logger")
            return ifNone(logger).get("level")
        except NullPointerException as e:
            return "info"
    
    def getLoggingFormat(self):
        try:
            application = self._config.get("application")
            logger = ifNone(application).get("logger")
            return ifNone(logger).get("format")
        except NullPointerException as e:
            return "%(levelname)s:%(asctime)s:%(message)s"
            
    def getLoggingFilePath(self):
        try:
            application = self._config.get("application")
            logger = ifNone(application).get("logger")
            return IfNone(logger).get("log_file_path")
        except NullPointerException as e:
            return "bingdog.log"
            
class ConfiguratorException(Exception):
    def __init__(self, confClass):
        super().__init__(confClass, " is not a class of Configurator.")
        
    